@@ -1,45 +1,64 @@
 import contextlib
 import io
+import json
 import logging
 import os
-import re
 import subprocess
 import time
 from contextlib import contextmanager
 from datetime import datetime, timedelta
+from json import JSONDecodeError
 from math import ceil
 from threading import Thread
 from typing import NamedTuple, Optional, Dict, List, Callable
 
+import cortex
+import requests
 import yaml
+from cortex.binary import get_cli_path
 from psycopg2._psycopg import DatabaseError
 from psycopg2.extras import NamedTupleCursor
 from psycopg2.pool import ThreadedConnectionPool
+from requests.adapters import HTTPAdapter, Retry
 
 from cortex_serving_client.command_line_wrapper import _verbose_command_wrapper
 from cortex_serving_client.deployment_failed import DeploymentFailed, COMPUTE_UNAVAILABLE_FAIL_TYPE, \
-    DEPLOYMENT_TIMEOUT_FAIL_TYPE, DEPLOYMENT_ERROR_FAIL_TYPE
+    DEPLOYMENT_TIMEOUT_FAIL_TYPE, DEPLOYMENT_ERROR_FAIL_TYPE, DEPLOYMENT_JOB_NOT_DEPLOYED_FAIL_TYPE
 from cortex_serving_client.printable_chars import remove_non_printable
 
 
+KIND_REALTIME_API = 'RealtimeAPI'
+KIND_BATCH_API = 'BatchAPI'
+KIND_DEFAULT = KIND_REALTIME_API
+
+NOT_DEPLOYED = "not deployed"
+
 """
-Details: https://www.cortex.dev/deployments/statuses
+Source: https://github.com/cortexlabs/cortex/blob/0.23/pkg/types/status/code.go
 On some places custom "not_deployed" status may be used.
 """
 
-LIVE_STATUS = "live"
-UPDATING_STATUS = "updating"
-COMPUTE_UNAVAILABLE_STATUS = "compute unavailable"
-CORTEX_STATUSES = [LIVE_STATUS, UPDATING_STATUS, "error", "error (out of memory)", COMPUTE_UNAVAILABLE_STATUS]  #
+LIVE_STATUS = "status_live"
+UPDATING_STATUS = "status_updating"
+COMPUTE_UNAVAILABLE_STATUS = "status_stalled"
+CORTEX_STATUSES = [LIVE_STATUS, UPDATING_STATUS, "status_error", "status_oom", COMPUTE_UNAVAILABLE_STATUS]  #
+NOT_DEPLOYED_STATUS = "status_not_deployed"
+
+JOB_STATUS_SUCCEEDED = 'status_succeeded'
+JOB_STATUS_ENQUEUING = 'status_enqueuing'
+JOB_STATUS_RUNNING = 'status_running'
+
 
 CORTEX_DELETE_TIMEOUT_SEC = 10 * 60
 CORTEX_DEPLOY_REPORTED_TIMEOUT_SEC = 60
 CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT = 20 * 60
 CORTEX_DEFAULT_API_TIMEOUT = CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT
+CORTEX_MIN_API_TIMEOUT_SEC = CORTEX_DELETE_TIMEOUT_SEC
 CORTEX_DEPLOY_RETRY_BASE_SLEEP_SEC = 5 * 60
 CORTEX_STATUS_CHECK_SLEEP_SEC = 10
 INFINITE_TIMEOUT_SEC = 30 * 365 * 24 * 60 * 60  # 30 years
 
+CORTEX_PATH = get_cli_path()
 
 logger = logging.getLogger('cortex_client')
 __cortex_client_instance = None
@@ -54,6 +73,8 @@ class CortexClient:
         self.db_connection_pool = db_connection_pool
         self._init_garbage_api_collector(gc_interval_sec)
         self.cortex_env = cortex_env
+        self.cortex_vanilla = cortex.client(self.cortex_env)
+        logger.info(f'Constructing CortexClient for {CORTEX_PATH}.')
 
     def deploy_single(
         self,
@@ -92,20 +113,19 @@ class CortexClient:
         """
 
         name = deployment["name"]
+        if "kind" not in deployment:
+            deployment["kind"] = KIND_DEFAULT
+
         predictor_yaml_str = yaml.dump([deployment], default_flow_style=False)
 
-        if api_timeout_sec < CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT:
-            logger.info(f'API timeout {api_timeout_sec} is be smaller than default deployment timeout {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}. Setting it to {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}.')
-            api_timeout_sec = CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT
+        if api_timeout_sec < CORTEX_MIN_API_TIMEOUT_SEC:
+            logger.info(f'API timeout {api_timeout_sec} is be smaller than minimal API timeout {CORTEX_MIN_API_TIMEOUT_SEC}. Setting it to {CORTEX_MIN_API_TIMEOUT_SEC} to avoid excessive GC.')
+            api_timeout_sec = CORTEX_MIN_API_TIMEOUT_SEC
 
         if deployment_timeout_sec < CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT:
-            logger.info(f'Deployment timeout {api_timeout_sec} is be smaller than default deployment timeout {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}. Setting it to {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}.')
+            logger.info(f'Deployment timeout {deployment_timeout_sec} is be smaller than default deployment timeout {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}. Setting it to {CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT}.')
             deployment_timeout_sec = CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT
 
-        if api_timeout_sec < deployment_timeout_sec:
-            logger.warning(f'API timeout for {name} of {api_timeout_sec} is shorter than deployment timeout of {deployment_timeout_sec}. This may cause unintended garbage collection! Setting API timeout to deployment timeout.')
-            api_timeout_sec = deployment_timeout_sec
-
         # file has to be created the dir to which python predictors are described in yaml
         filename = f"{name}.yaml"
         filepath = f"{dir}/{filename}"
@@ -114,21 +134,12 @@ class CortexClient:
             try:
                 self._collect_garbage()
                 with str_to_public_temp_file(predictor_yaml_str, filepath):
-                    gc_timeout_sec = api_timeout_sec + CORTEX_DELETE_TIMEOUT_SEC
-                    logger.info(f"Deployment {name} has deployment timeout: {deployment_timeout_sec} sec, garbage collection timeout: {gc_timeout_sec} seconds.")
-                    with self._open_db_cursor() as cursor:
-                        cursor.execute(
-                            """
-                            insert into cortex_api_timeout(api_name, ultimate_timeout) values (%s, transaction_timestamp() + %s * interval '1 second')
-                            on conflict (api_name) do update
-                                set ultimate_timeout = transaction_timestamp() + %s * interval '1 second', modified = transaction_timestamp()
-                        """,
-                            [name, gc_timeout_sec, gc_timeout_sec],
-                        )
+                    gc_timeout_sec = deployment_timeout_sec + CORTEX_DELETE_TIMEOUT_SEC
+                    logger.info(f"Deployment {name} has deployment timeout {deployment_timeout_sec}sec and GC timeout {gc_timeout_sec}sec.")
+                    self._insert_or_update_gc_timeout(name, gc_timeout_sec)
+                    _verbose_command_wrapper([CORTEX_PATH, "deploy", filename, f"--env={self.cortex_env}", "--yes", "-o=json"], cwd=dir)
 
-                    _verbose_command_wrapper(["cortex", "deploy", filename, f"--env={self.cortex_env}"], cwd=dir)
-
-                if print_logs:
+                if print_logs and deployment['kind'] == KIND_REALTIME_API:
                     self._cortex_logs_print_async(name)
 
                 start_time = time.time()
@@ -136,6 +147,9 @@ class CortexClient:
                     get_result = self.get(name)
                     time_since_start = ceil(time.time() - start_time)
                     if get_result.status == LIVE_STATUS:
+                        gc_timeout_sec = api_timeout_sec + CORTEX_DELETE_TIMEOUT_SEC
+                        logger.info(f"Deployment {name} successful. Setting API timeout {api_timeout_sec}sec and GC timeout {gc_timeout_sec}sec.")
+                        self._insert_or_update_gc_timeout(name, gc_timeout_sec)
                         return get_result
 
                     elif get_result.status == COMPUTE_UNAVAILABLE_STATUS:
@@ -154,7 +168,7 @@ class CortexClient:
                         raise DeploymentFailed(f"Deployment of {name} timed out after {deployment_timeout_sec} secs.",
                                                DEPLOYMENT_TIMEOUT_FAIL_TYPE, name, time_since_start)
 
-                    logger.info(f"Sleeping during deployment of {name} until next status check. Current get_result: {get_result}.")
+                    logger.info(f"Sleeping during deployment of {name} until next status check. Current status: {get_result.status}.")
                     time.sleep(CORTEX_STATUS_CHECK_SLEEP_SEC)
 
             except DeploymentFailed as e:
@@ -212,6 +226,61 @@ class CortexClient:
         finally:
             self.delete(deployment["name"])
 
+    def deploy_batch_api_and_run_job(self,
+                                     deployment,
+                                     job_spec,
+                                     dir,
+                                     deployment_timeout_sec=CORTEX_DEFAULT_DEPLOYMENT_TIMEOUT,
+                                     api_timeout_sec=CORTEX_DEFAULT_API_TIMEOUT,
+                                     print_logs=False,
+                                     n_retries=0) -> "CortexGetResult":
+
+        with self.deploy_temporarily(
+                deployment,
+                dir=dir,
+                deployment_timeout_sec=deployment_timeout_sec,
+                api_timeout_sec=api_timeout_sec,
+                print_logs=print_logs,
+                n_retries=n_retries
+        ) as get_result:
+
+            logger.info(f'BatchAPI {deployment["name"]} deployed. Starting a job.')
+            s = requests.Session()
+            http_adapter = HTTPAdapter(max_retries=Retry(total=5, backoff_factor=3))
+            s.mount('http://', http_adapter)
+            s.mount('https://', http_adapter)
+            job_id = None
+            for retry in range(3):
+                job_json = s.post(get_result.endpoint, json=job_spec, timeout=10 * 60).json()
+                if 'job_id' in job_json:
+                    job_id = job_json['job_id']
+                    break
+
+                else:
+                    time.sleep(3 * 2**retry)
+                    continue
+
+            if job_id is None:
+                raise ValueError(f'job_id not in job_json {json.dumps(job_json)}')
+
+            if print_logs:
+                self._cortex_logs_print_async(deployment['name'], job_id)
+
+            time.sleep(30)  # Don't call too early: https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
+
+            job_status = JOB_STATUS_ENQUEUING
+            while job_status in (JOB_STATUS_ENQUEUING, JOB_STATUS_RUNNING):
+                job_result = self.get(deployment["name"], job_id)
+                job_status = job_result.status
+                time.sleep(30)
+
+            if job_status == NOT_DEPLOYED_STATUS:
+                # TODO Better retry mechanism could be used https://gitter.im/cortexlabs/cortex?at=5f7fe4c01cbba72b63cb745f
+                raise DeploymentFailed('Job unexpectedly undeployed.', DEPLOYMENT_JOB_NOT_DEPLOYED_FAIL_TYPE, deployment['name'], -1)
+
+            logger.info(f'BatchAPI {deployment["name"]} job {job_id} ended with status {job_status}. Deleting the BatchAPI.')
+            return job_result
+
     def postpone_api_timeout(self, name: str, timeout_timestamp: datetime):
         ultimate_timeout = timeout_timestamp + timedelta(seconds=CORTEX_DELETE_TIMEOUT_SEC)
         with self._open_db_cursor() as cursor:
@@ -222,49 +291,55 @@ class CortexClient:
 
     def raise_on_cluster_down(self):
         try:
-            input_in_case_endpoint_prompt = b"n\n"
-            _verbose_command_wrapper(["cortex", "get", f"--env={self.cortex_env}"], input=input_in_case_endpoint_prompt)
+            self.get_all()
 
-        except ValueError as e:
-            raise ValueError("Cluster is likely down! Check the exception text") from e
+        except (ValueError, JSONDecodeError) as e:
+            raise ValueError(f"Cluster is likely down: {e}.") from e
+
+    def get(self, name, job_id=None) -> "CortexGetResult":
+        job_id_or_empty = [job_id] if job_id is not None else []
+        cmd = [CORTEX_PATH, "get", name] + job_id_or_empty + [f"--env={self.cortex_env}", "-o=json"]
+        out = _verbose_command_wrapper(cmd, allow_non_0_return_code_on_stdout_sub_strs=[NOT_DEPLOYED])
+        try:
+            json_dict = json.loads(out.strip())
+
+        except JSONDecodeError as e:
+            logger.debug(f'Encountered {e} but ignoring.')
+            json_dict = None
 
-    def get(self, name) -> "CortexGetResult":
-        out = _verbose_command_wrapper(["cortex", "get", name, f"--env={self.cortex_env}"])
         lines = out.splitlines()
-        if lines[1].split(" ")[0] == "status":
-            return self._parse_get_deployed(out)
+        first_line = lines[0] if len(lines) > 0 else ''
+        if json_dict is not None:
+            if job_id is not None:
+                job_status = json_dict['job_status']["status"]
+                return CortexGetResult(job_status, None, json_dict)
 
-        elif "not deployed" in lines[1]:
-            return CortexGetResult("not_deployed", None)
+            first_json_dict = json_dict[0]
+            if first_json_dict['spec']['kind'] == KIND_BATCH_API:
+                # BatchAPI has no status
+                return CortexGetResult(LIVE_STATUS, first_json_dict['endpoint'], first_json_dict)
+
+            else:
+                return CortexGetResult(first_json_dict['status']['status_code'], first_json_dict['endpoint'], first_json_dict)
+
+        elif NOT_DEPLOYED in first_line:
+            return CortexGetResult(NOT_DEPLOYED_STATUS, None, dict())
 
         else:
-            raise ValueError(f"For api: {name} got unsupported Cortex output:\n{out}")
-
-    @staticmethod
-    def _parse_get_deployed(cmd_out: str):
-        lines = cmd_out.splitlines()
-        status = lines[2].split("  ")[0].strip()
-        if status == LIVE_STATUS:
-            for line in lines:
-                line_split = line.split(" ")
-                if line_split[0] == "endpoint:":
-                    url = line_split[1]
-                    return CortexGetResult(status, url)
-
-            raise ValueError(f"Unsupported Cortex output:\n{cmd_out}")
-
-        return CortexGetResult(status, None)
+            raise ValueError(f"For api {name} with job_id {job_id} got unsupported Cortex output:\n{out}")
 
     def get_all(self) -> List["CortexGetAllStatus"]:
-        out = _verbose_command_wrapper(["cortex", "get", f"--env={self.cortex_env}"])
-        return cortex_parse_get_all(out)
+        input_in_case_endpoint_prompt = b"n\n"
+        out = _verbose_command_wrapper([CORTEX_PATH, "get", f"--env={self.cortex_env}", "-o=json"], input=input_in_case_endpoint_prompt)
+        json_dict = json.loads(out.strip())
+        return [CortexGetAllStatus(e['spec']['name'], e['status']['status_code'] if e['spec']['kind'] == KIND_REALTIME_API else LIVE_STATUS, e) for e in json_dict]
 
     def delete(self, name, force=False, timeout_sec=CORTEX_DELETE_TIMEOUT_SEC, cursor=None) -> "CortexGetResult":
         """
         Executes delete and checks that the API was deleted. If not tries force delete. If that fails, raises exception.
         """
 
-        delete_cmd = ["cortex", "delete", name, f"--env={self.cortex_env}"]
+        delete_cmd = [CORTEX_PATH, "delete", name, f"--env={self.cortex_env}", "-o=json"]
         if force:
             delete_cmd.append("-f")
 
@@ -274,12 +349,12 @@ class CortexClient:
             start_time = time.time()
             while True:
                 get_result = self.get(name)
-                if get_result.status == "not_deployed":
+                if get_result.status == NOT_DEPLOYED_STATUS:
                     return get_result
 
                 else:
                     _verbose_command_wrapper(delete_cmd, input=accept_deletion_if_asked_input,
-                                             allow_non_0_return_code_on_stdout_sub_strs=['not deployed'])
+                                             allow_non_0_return_code_on_stdout_sub_strs=[NOT_DEPLOYED])
 
                 if start_time + timeout_sec < time.time():
                     if force:
@@ -289,22 +364,28 @@ class CortexClient:
                         logger.error(f'Timeout of delete cmd {delete_cmd} after {timeout_sec} seconds. Will attempt to force delete now.')
                         return self.delete(name, force=True, timeout_sec=timeout_sec, cursor=cursor)
 
-                logger.info(f"During delete of {name} sleeping until next status check. Current get_result: {get_result}.")
+                logger.info(f"During delete of {name} sleeping until next status check. Current result: {get_result.status}.")
                 time.sleep(CORTEX_STATUS_CHECK_SLEEP_SEC)
 
         finally:
             self._open_cursor_if_none(cursor, self._del_db_api_row, name)
 
-    def _cortex_logs_print_async(self, name):
-        def listen_on_logs():
-            with subprocess.Popen(["cortex", "logs", name, f"--env={self.cortex_env}"], stdout=subprocess.PIPE) as logs_sp:
+    def _cortex_logs_print_async(self, name, job_id=None):
+        def listen_on_logs(cmd_arr):
+            with subprocess.Popen(cmd_arr, stdout=subprocess.PIPE) as logs_sp:
                 with io.TextIOWrapper(logs_sp.stdout, encoding="utf-8") as logs_out:
                     for line in logs_out:
                         print_line = remove_non_printable(line.rstrip('\n'))
                         if len(print_line) > 0:
                             logger.info(print_line)
 
-        worker = Thread(target=listen_on_logs, daemon=True, name=f'api_{name}')
+        if job_id is None:
+            cmd_arr = [CORTEX_PATH, "logs", name, f"--env={self.cortex_env}"]
+
+        else:
+            cmd_arr = [CORTEX_PATH, "logs", name, job_id, f"--env={self.cortex_env}"]
+
+        worker = Thread(target=listen_on_logs, args=(cmd_arr,), daemon=True, name=f'api_{name}')
         worker.start()
 
     @staticmethod
@@ -321,6 +402,17 @@ class CortexClient:
         except DatabaseError:
             logger.warning(f"Ignoring exception during cortex_api_timeout record for {name} modifying ultimate_timeout to delete the api", exc_info=True)
 
+    def _insert_or_update_gc_timeout(self, name: str, gc_timeout_seconds_from_now: float):
+        with self._open_db_cursor() as cursor:
+            cursor.execute(
+                """
+                insert into cortex_api_timeout(api_name, ultimate_timeout) values (%s, transaction_timestamp() + %s * interval '1 second')
+                on conflict (api_name) do update
+                    set ultimate_timeout = transaction_timestamp() + %s * interval '1 second', modified = transaction_timestamp()
+            """,
+                [name, gc_timeout_seconds_from_now, gc_timeout_seconds_from_now],
+            )
+
     @staticmethod
     def _del_db_api_row(cur, name):
         try:
@@ -350,7 +442,7 @@ class CortexClient:
                 )
                 for api_row in cur.fetchall():
                     api_name = api_row.api_name
-                    if self.get(api_name).status != "not_deployed":
+                    if self.get(api_name).status != NOT_DEPLOYED_STATUS:
                         logger.warning(f"Collecting Cortex garbage - timed-out deployed API: {api_name}")
                         self.delete(api_name, cursor=cur)
 
@@ -434,30 +526,13 @@ def open_pg_cursor(db_connection_pool, key=None):
 class CortexGetAllStatus(NamedTuple):
     api: str
     status: str
-
-
-def cortex_parse_get_all(out: str) -> List[CortexGetAllStatus]:
-    if "no apis are deployed" in out:
-        return []
-
-    lines = out.splitlines()
-
-    header_split = re.split(r"\s+", lines[1])
-    if header_split[0] == "api" and header_split[1] == "status":
-        result = []
-        for row_line in lines[2:]:
-            row = re.split(r"\s{2,}", row_line)
-            result.append(CortexGetAllStatus(row[0].strip(), row[1].strip()))
-
-        return result
-
-    else:
-        raise ValueError(f"Cannot parse output:\n{out}")
+    response: dict
 
 
 class CortexGetResult(NamedTuple):
     status: str
     endpoint: Optional[str]
+    response: dict
 
 
 def file_to_str(path: str) -> str:
